#!/usr/bin/env perl

package GoIP;

use strict;

#use File::Lockfile;

use IO::Select;
use IO::Socket::INET;
use Data::Dumper;
use Timer::Simple;

sub new_id($) {
    my $self = shift;

    $self->{id}++;

    $self->{id} = 1 if $self->{id} > 10000;

    return $self->{id};
}

sub send($$) {
    my ( $self, $buf ) = @_;

    print "> " . $buf . "\n";

    $self->{sock}->send( $buf, MSG_NOSIGNAL );

}

sub recv_req($$) {
    my ( $self, $buf ) = @_;

    my @msg = split( /;/, $buf );

    return unless $msg[0] =~ /^req:([-+\d]+)/;
    my $id = $1;
    return unless $msg[1] eq 'id:' . $self->{goip_id};
    return unless $msg[2] eq 'pass:' . $self->{goip_password};

    my $reply = 'reg:' . $id . ';status:0;';

    $self->send($reply);

}

sub recv_password($$) {
    my ( $self, $buf ) = @_;

    return unless $buf =~ /^PASSWORD (\d+)/;

    my $id = $1;

    my $reply = 'PASSWORD ' . $id . ' ' . $self->{goip_password} . "\n";

    $self->send($reply);

}

sub recv_error($$) {
    my ( $self, $buf ) = @_;

    return unless $buf =~ /^ERROR (\d+)/;

    my $id = $1;

    $self->{error}->{$id} = 1;

}

sub recv_done($$) {
    my ( $self, $buf ) = @_;

    return unless $buf =~ /^DONE (\d+)/;

    my $id = $1;

    $self->{done}->{$id} = 1;

}

sub recv_ok($$) {
    my ( $self, $buf ) = @_;

    return unless $buf =~ /^OK (\d+)/;

    my $id = $1;

    $self->{ok}->{$id} = 1;

}

sub recv_send($$) {
    my ( $self, $buf ) = @_;

    return unless $buf =~ /^SEND (\d+)/;

    my $id = $1;

    $self->{send}->{$id} = 1;
}

sub recv_receive($$) {

    my ( $self, $buf ) = @_;

    my $sms = {};

    my @msg = split( /;/, $buf );

    return unless $msg[0] =~ /^RECEIVE:([-+\d]+)/;
    my $id = $1;

    return unless $msg[1] eq 'id:' . $self->{goip_id};
    return unless $msg[2] eq 'password:' . $self->{goip_password};
    return unless $msg[3] =~ /^srcnum:(.*)/;

    $sms->{origin} = $1;

    return
      unless $buf =~
          /^RECEIVE:[^;]+;id:[^;]+;password:[^;]+;srcnum:[^;]+;msg:(.*)$/;

    $sms->{body} = $1;

    my $reply = 'RECEIVE ' . $id . ' OK' . "\n";

    $self->send($reply);

    push @{ $self->{received} }, $sms;

    print "SMS received from " . $sms->{origin}, " message ", $sms->{body},
      "\n";

}

sub handle_message($$) {
    my ( $self, $buf ) = @_;

    print "< ", $buf, "\n";

    if ( $buf =~ /^req:/ ) {
        $self->recv_req($buf);
    }
    elsif ( $buf =~ /^PASSWORD/ ) {
        $self->recv_password($buf);
    }
    elsif ( $buf =~ /^ERROR/ ) {
        $self->recv_error($buf);
    }
    elsif ( $buf =~ /^SEND/ ) {
        $self->recv_send($buf);
    }
    elsif ( $buf =~ /^OK/ ) {
        $self->recv_ok($buf);
    }
    elsif ( $buf =~ /^DONE/ ) {
        $self->recv_done($buf);
    }
    elsif ( $buf =~ /^RECEIVE/ ) {
        $self->recv_receive($buf);
    }
    else {
        print $buf;
    }

}

sub dispatch($$) {
    my ( $self, $timeout ) = @_;

    my @ready = $self->{select}->can_read($timeout);

    return undef unless scalar(@ready) > 0;

    my $buf = undef;
    my $from = $self->{sock}->recv( $buf, 1024, MSG_DONTWAIT );

    return undef unless defined $from;

    my ( $port, $ipaddr ) = sockaddr_in($from);

    return 1 if length($buf) == 0;

    $self->handle_message($buf);

    return 1;

}

sub done($$) {
    my ( $self, $id ) = @_;

    my $timer        = Timer::Simple->new();
    my $resend_timer = Timer::Simple->new();

    my $done = 'DONE ' . $id . "\n";

    $self->send($done);

    until ( exists $self->{done}->{$id} ) {
        $self->dispatch(1);
        if ( $resend_timer->elapsed >= 1 ) {
            $self->send($done);
            $resend_timer = Timer::Simple->new();
        }

        return if $timer->elapsed > 5;
    }

    delete $self->{done}->{$id};
}

sub sms($$$) {
    my ( $self, $dest, $msg ) = @_;

    my $id = $self->new_id;

    my $timer        = Timer::Simple->new();
    my $resend_timer = Timer::Simple->new();

    my $msg = "MSG " . $id . " " . length($msg) . " " . $msg . "\n";

    $self->send($msg);

    delete $self->{send}->{$id};
    delete $self->{error}->{$id};

    until ( exists $self->{send}->{$id} ) {

        $self->dispatch(1);

        if ( $resend_timer->elapsed >= 1 ) {
            $self->send($msg);
            $resend_timer = Timer::Simple->new();
        }

        if ( exists $self->{error}->{$id} ) {
            delete $self->{error}->{$id};
            return undef;
        }

        return undef if $timer->elapsed > 5;

    }

    delete $self->{send}->{$id};

    my $telid = $self->new_id;

    my $submit = 'SEND ' . $id . ' ' . $telid . ' ' . $dest . "\n";

    $self->send($submit);

    $timer        = Timer::Simple->new();
    $resend_timer = Timer::Simple->new();

    until ( exists $self->{ok}->{$id} ) {

        $self->dispatch(1);

        if ( $resend_timer->elapsed >= 1 ) {
            $self->send($submit);
            $resend_timer = Timer::Simple->new();
        }

        if ( exists $self->{error}->{$id} ) {
            delete $self->{error}->{$id};
            $self->done($id);
            return undef;
        }

        return undef if $timer->elapsed > 30;
    }
    delete $self->{ok}->{$id};

    $self->done($id);

    print "Sending a success\n";
    return 1;

}

sub new ($;$) {
    my ( $class, $parm ) = @_;
    my $self;

    $self->{goip_port}     = $parm->{goip_port}     || 9991;
    $self->{goip_addr}     = $parm->{goip_addr}     || '10.32.87.22';
    $self->{local_addr}    = $parm->{local_addr}    || '10.32.87.15';
    $self->{local_port}    = $parm->{local_port}    || 44444;
    $self->{goip_id}       = $parm->{goip_id}       || 'goip';
    $self->{goip_password} = $parm->{goip_password} || 'goip';

    $self->{error}    = {};
    $self->{send}     = {};
    $self->{ok}       = {};
    $self->{done}     = {};
    $self->{received} = [];

    $self->{id} = int( rand(10000) );

#    $self->{lockfile}=File::Lockfile->new('goip.pid','/var/lock/goip');
#    while ( my $pid = $self->{lockfile}->check ) {
#	sleep(2);
#    }
#    $self->{lockfile}->write


    do {
  
    sleep(1);

    $self->{sock} = IO::Socket::INET->new(
        PeerPort  => $self->{goip_port},
        PeerAddr  => $self->{goip_addr},
        Proto     => 'udp',
        LocalAddr => $self->{local_addr},
        LocalPort => $self->{local_port},
        ReuseAddr => 1,
        Blocking  => 0
    );

    } until defined ($self->{sock});

    $self->{select} = IO::Select->new();

    $self->{select}->add( $self->{sock} );

    return bless $self, $class;

}

1;
