package SMS;

use strict;
use Encode qw/encode decode/;
use Data::Dumper;
use DateTime;
use Digest::MD5 qw/md5/;

my @message_direction = ( "SC->ME", "ME->SC" );
my @message_type = (
    "None",          "Cell Broadcast", "Deliver", "Deliver Report",
    "Submit Report", "Submit",         "Command", "Status Report"
);
my @address_type = (
    "unknown",   "international", "national",    "network specific",
    "subsciber", "alphanumeric",  "abbreviated", "reserved"
);
my @address_plan = (
    "unknown", "isdn/telephone", "data",     "telex",
    "plan 5",  "plan 6",         "national", "private",
    "ermes",   "reserved"
);

my @validity_type = ( "absent", "reserved", "relative", "absolute" );

my %gsm_status = (
    0  => 'complete: Received OK',
    1  => 'complete: Unable to Confirm',
    2  => 'complete: Replaced',
    32 => 'temporary: Conjestion',
    33 => 'temporary: ME busy',
    34 => 'temporary: No response from ME',
    35 => 'temporary: Service Rejected',
    36 => 'temporary: QoS not available',
    37 => 'temporary: Error in ME',
    64 => 'error: remote procedure error',
    65 => 'error: incompatible destination',
    66 => 'error: rejected my ME',
    67 => 'error: not obtainable',
    68 => 'error: QoS not available',
    69 => 'error: No interworking available',
    70 => 'error: Validity Period Expired',
    71 => 'error: Deleted by originator',
    72 => 'error: Deleted by SC admin',
    73 => 'error: Message lost',
);

use constant {

    GSM_VALIDITYPERIODTYPE_ABSENT   => 0,
    GSM_VALIDITYPERIODTYPE_RELATIVE => 2,
    GSM_VALIDITYPERIODTYPE_ABSOLUTE => 3,

    GSM_ADDR_TYPE_UNKNOWN       => 0,
    GSM_ADDR_TYPE_INTERNATIONAL => 1,
    GSM_ADDR_TYPE_NATIONAL      => 2,
    GSM_ADDR_TYPE_ALPHANUMERIC  => 5,

    GSM_ADDR_PLAN_ISDN_TELEPHONE => 1,

    SC_TO_ME => 0,
    ME_TO_SC => 1,

    #SC->ME
    GSM_MTI_DELIVER       => 0,
    GSM_MTI_SUBMIT_REPORT => 1,
    GSM_MTI_STATUS_REPORT => 2,

    #ME->SC
    GSM_MTI_DELIVER_REPORT => 0,
    GSM_MTI_SUBMIT         => 1,
    GSM_MTI_COMMAND        => 2,

    MT_NONE           => 0,
    MT_CELL_BROADCAST => 1,
    MT_DELIVER        => 2,
    MT_DELIVER_REPORT => 3,
    MT_SUBMIT_REPORT  => 4,
    MT_SUBMIT         => 5,
    MT_STATUS_REPORT  => 6,
    MT_COMMAND        => 7,

    GSM_IE_CONCAT_8BIT  => 0x00,
    GSM_IE_APAS_8BIT    => 0x04,
    GSM_IE_APAS_16BIT   => 0x04,
    GSM_IE_CONCAT_16BIT => 0x08,

    GSM_DCS_STANDARD_GSM => 0x00

};

sub unpack_7bit_string($$) {
    my ( $i, $l ) = @_;
    my $o = "";

    my $b = "";

    for my $v (@$i) {
        $b .= unpack( 'b8', pack( 'C', $v ) );
    }

    while ( $b =~ /^(.......)(.*)$/ ) {
        $b = $2;
        $o .= pack( 'b7', $1 );
        last unless --$l;
    }

    return $o;
}

sub pack_7bit_string($) {
    my $i = shift;
    my $o = [];

    my $b = "";

    for my $c ( split //, $i ) {
        $b .= unpack( 'b7', $c );
    }

    while ( length($b) % 8 ) {
        $b .= "0";
    }

    while ( $b =~ /^(........)(.*)$/ ) {
        $b = $2;
        push @$o, unpack( 'C', pack( 'b8', $1 ) );
    }

    return $o;
}

sub remove_n_bits($$) {
    my ( $i, $n ) = @_;
    my $o = [];

    my $b = "";

    for my $v (@$i) {
        $b .= unpack( 'b8', pack( 'C', $v ) );
    }

    $b = substr( $b, $n );

    while ( length($b) % 8 ) {
        $b .= "0";
    }

    while ( $b =~ /^(........)(.*)$/ ) {
        $b = $2;
        push @$o, unpack( 'C', pack( 'b8', $1 ) );
    }

    return $o;
}

sub add_n_bits($$) {
    my ( $i, $n ) = @_;
    my $o = [];

    my $b = "";

    for my $v (@$i) {
        $b .= unpack( 'b8', pack( 'C', $v ) );
    }

    while ( $n-- ) {
        $b = "0" . $b;
    }

    while ( length($b) % 8 ) {
        $b .= "0";
    }

    while ( $b =~ /^(........)(.*)$/ ) {
        $b = $2;
        push @$o, unpack( 'C', pack( 'b8', $1 ) );
    }

    return $o;
}

sub nibble_to_digit($) {
    my $n = shift;
    return chr( ord('0') + $n ) if ( $n >= 0 ) and ( $n < 10 );
    return "";
}

sub parse_nibbles($) {
    my $pdu = shift;
    my $ret = "";

    my $b;

    while ( defined( $b = shift @$pdu ) ) {
        $ret .= nibble_to_digit( $b & 0xf );
        $ret .= nibble_to_digit( ( $b & 0xf0 ) >> 4 );
    }
    return $ret;
}

sub digit_to_nibble($) {
    my $n = shift;
    return 0xf unless $n =~ /^[0-9]$/;
    return ord($n) - ord('0');
}

sub encode_nibbles($) {
    my $n   = shift;
    my $ret = [];

    while ( $n =~ /^(.)(.)(.*)$/ ) {
        my $b = 0;
        $b |= nibble_to_digit($1);
        $b |= nibble_to_digit($2) << 4;
        $n = $3;
        push @$ret, $b;
    }

    if ( length($n) > 0 ) {
        $b = nibble_to_digit($n) | 0xf0;
        push @$ret, $b;
    }

    return $ret;
}

sub parse_address_text($) {
    my $num = shift;

    my $s = unpack_7bit_string( $num, 0 );
    $s = decode( "gsm0338", $s );

    return $s;
}

sub parse_address($$) {
    my ( $pdu, $l ) = @_;
    my $b;

    my $ret = {};

    if ( $l == 0 ) {
        $l = shift @{$pdu};
        $l = $l + 1;
        $l = $l >> 1;
        $l += 1;
    }

    $b = shift @{$pdu};
    return undef unless $b & 0x80;

    $ret->{type_id} = ( $b & 0x70 ) >> 4;
    $ret->{type}    = $address_type[ $ret->{type_id} ];
    $ret->{plan_id} = $b & 0xf;
    $ret->{plan}    = $address_plan[ $ret->{plan_id} ];

    $l--;

    my $num = [ splice @{$pdu}, 0, $l ];

    $ret->{raw} = [@$num];

    if ( $ret->{type_id} == GSM_ADDR_TYPE_UNKNOWN ) {
        $ret->{addr} = parse_nibbles($num);

	$ret->{addr} =~ s/^011/+/;
	$ret->{addr} =~ s/^00/+/;
	$ret->{addr} =~ s/^0/+44/;
	$ret->{addr} =~ s/^1/+1/;
    }
    elsif ( $ret->{type_id} == GSM_ADDR_TYPE_INTERNATIONAL ) {
        $ret->{addr} = "+" . parse_nibbles($num);
    }
    elsif ( $ret->{type_id} == GSM_ADDR_TYPE_NATIONAL ) {
        $ret->{addr} = parse_nibbles($num);
    }
    elsif ( $ret->{type_id} == GSM_ADDR_TYPE_ALPHANUMERIC ) {
        $ret->{addr} = parse_address_text($num);
    }
    else {
        $ret->{addr} = "unknown type";
    }

    return $ret;
}

sub encode_address($) {
    my $a   = shift;
    my $ret = [];

    if ( $a =~ /^\+([0-9]*)$/ ) {
        $a = $1;

        my $l = length($a);
        push @$ret, $l;

        $b = 0x80;
        $b |= GSM_ADDR_TYPE_INTERNATIONAL << 4;
        $b |= GSM_ADDR_PLAN_ISDN_TELEPHONE;

        push @$ret, $b;

        my $n = encode_nibbles($a);

        push @$ret, (@$n);
    }
    elsif ( $a =~ /^([0-9]*)$/ ) {
        $a = $1;

        my $l = length($a);
        push @$ret, $l;

        $b = 0x80;
        $b |= GSM_ADDR_TYPE_UNKNOWN << 4;
        $b |= GSM_ADDR_PLAN_ISDN_TELEPHONE;

        push @$ret, $b;

        my $n = encode_nibbles($a);

        push @$ret, (@$n);
    }

    return $ret;
}

sub relative_validity_to_secs ($) {
    my $v = shift;
    if ( $v < 144 ) {
        return ( $v * ( 5 * 60 ) );
    }
    if ( $v < 168 ) {
        $v -= 143;    # Half hours over 12 hours
        $v += 24;
        $v *= ( 30 * 60 );
        return $v;
    }
    if ( $v < 199 ) {
        $v -= 167;    # Days over one day
        $v++;
        $v *= ( 24 * 60 * 60 );
        return $v;
    }
    $v -= 199;        # weeks over one week
    $v++;
    $v *= ( 7 * 24 * 60 * 60 );
    return $v;
}

sub parse_timestamp($) {
    my $pdu = shift;
    my $ret = {};

    my $ts = [ splice @{$pdu}, 0, 7 ];
    my $s;

    $ret->{raw} = [@$ts];

    $s = parse_nibbles($ts);

    $ret->{time} = DateTime->new(
        year   => 2000 + int( substr( $s, 0,  2 ) ),
        month  => int( substr( $s,        2,  2 ) ),
        day    => int( substr( $s,        4,  2 ) ),
        hour   => int( substr( $s,        6,  2 ) ),
        minute => int( substr( $s,        8,  2 ) ),
        second => int( substr( $s,        10, 2 ) )
    );

    return $ret;
}

sub parse_validity($$) {
    my $ret = {};
    my ( $type_id, $pdu ) = @_;

    $ret->{type_id} = $type_id;
    $ret->{type}    = $validity_type[ $ret->{type_id} ];

    if ( $ret->{type_id} == GSM_VALIDITYPERIODTYPE_RELATIVE ) {
        $ret->{relative}->{raw_value} = shift @$pdu;
        $ret->{relative}->{secs} =
          relative_validity_to_secs( $ret->{relative}->{raw_value} );
    }
    if ( $ret->{type_id} == GSM_VALIDITYPERIODTYPE_ABSOLUTE ) {
        my $ts = [ splice @{$pdu}, 0, 7 ];
        $ret->{absolute} = parse_timestamp($ts);
    }

    return $ret;
}

sub parse_user_header($) {
    my $udh = shift;
    my $ret = {};
    my $hn  = 0;

    my $ie_type;

    while ( defined( $ie_type = shift @$udh ) ) {
        my $cur = {};
        $cur->{ie_type} = $ie_type;
        $cur->{ie_len}  = shift @$udh;

        my @ie = ( splice @{$udh}, 0, $cur->{ie_len} );

        if ( $ie_type == GSM_IE_CONCAT_8BIT ) {
            $cur->{concat}           = 1;
            $cur->{concat_reference} = $ie[0];
            $cur->{concat_total}     = $ie[1];
            $cur->{concat_sequence}  = $ie[2];
            $ret->{concat}           = 1;
            $ret->{concat_reference} = $ie[0];
            $ret->{concat_total}     = $ie[1];
            $ret->{concat_sequence}  = $ie[2];
        }
        elsif ( $ie_type = GSM_IE_CONCAT_16BIT ) {
            $cur->{concat}           = 1;
            $cur->{concat_reference} = $ie[0] + ( $ie[1] << 8 );
            $cur->{concat_total}     = $ie[2];
            $cur->{concat_sequence}  = $ie[3];
            $ret->{concat}           = 1;
            $ret->{concat_reference} = $ie[0] + ( $ie[1] << 8 );
            $ret->{concat_total}     = $ie[2];
            $ret->{concat_sequence}  = $ie[3];
        }

        $cur->{raw} = [@ie];

        $ret->{ $hn++ } = $cur;
    }

    $ret->{nh} = $hn;

    return $ret;
}

sub parse_user_data($$$$) {
    my ( $pid, $dcs, $ud, $udl ) = @_;
    my $ret = {};
    my $s;

    $ret->{raw} = [@$ud];

    if ( $dcs == GSM_DCS_STANDARD_GSM || ( ( $dcs & 0xf4 ) == 0xf0 ) ) {

        $s = unpack_7bit_string( $ud, $udl );
        $s = decode( "gsm0338", $s );

    }
    else {
        $s = pack( 'C*', @$ud );

    }

    $ret->{text} = $s;

    return $ret;
}

sub parse_deliver($$) {
    my ( $ret, $pdu ) = @_;

    $ret->{originator} = parse_address( $pdu, 0 );

    $ret->{pid} = shift @$pdu;
    $ret->{dcs} = shift @$pdu;

    $ret->{sc_stamp} = parse_timestamp($pdu);

    my $udl = shift @$pdu;
    my $udl_bytes;
    my $slop = 0;

    if ( $ret->{dcs} == GSM_DCS_STANDARD_GSM ) {
        $udl_bytes = ( $udl * 7 + 7 ) >> 3;
    }
    else {
        $udl_bytes = $udl;
    }

    if ( $ret->{user_data_header_indicator} == 1 ) {
        my $udhl = shift(@$pdu);

        $udl_bytes -= $udhl + 1;

        if ( $ret->{dcs} == GSM_DCS_STANDARD_GSM ) {
            my $bits_lost = 8 * ( $udhl + 1 );

            $slop = $bits_lost % 7;

            $slop = 7 - $slop;
            $udl -= ( $bits_lost + $slop ) / 7;

        }
        else {
            $udl -= $udhl + 1;
            $slop = 0;
        }

        my $udh = [ splice @{$pdu}, 0, $udhl ];

        $ret->{user_header} = parse_user_header($udh);
    }

    my $ud = [ splice @{$pdu}, 0, $udl_bytes ];

    $ud = remove_n_bits( $ud, $slop ) if $slop > 0;

    $ret->{user_data} = parse_user_data( $ret->{pid}, $ret->{dcs}, $ud, $udl );

}

sub parse_submit($$) {
    my ( $ret, $pdu ) = @_;

    $ret->{reference} = shift @$pdu;
    $ret->{destination} = parse_address( $pdu, 0 );

    $ret->{pid} = shift @$pdu;
    $ret->{dcs} = shift @$pdu;

    $ret->{validity} = parse_validity( $ret->{validity_period_type_id}, $pdu );

    my $udl = shift @$pdu;
    my $udl_bytes;
    my $slop = 0;

    if ( $ret->{dcs} == GSM_DCS_STANDARD_GSM ) {
        $udl_bytes = ( $udl * 7 + 7 ) >> 3;
    }
    else {
        $udl_bytes = $udl;
    }

    if ( $ret->{user_data_header_indicator} == 1 ) {
        my $udhl = shift(@$pdu);

        $udl_bytes -= $udhl + 1;

        if ( $ret->{dcs} == GSM_DCS_STANDARD_GSM ) {
            my $bits_lost = 8 * ( $udhl + 1 );

            $slop = $bits_lost % 7;

            $slop = 7 - $slop;

            $udl -= ( $bits_lost + $slop ) / 7;

        }
        else {
            $udl -= $udhl + 1;
            $slop = 0;
        }

        my $udh = [ splice @{$pdu}, 0, $udhl ];

        $ret->{user_header} = parse_user_header($udh);
    }

    my $ud = [ splice @{$pdu}, 0, $udl_bytes ];

    $ud = remove_n_bits( $ud, $slop ) if $slop > 0;

    $ret->{user_data} = parse_user_data( $ret->{pid}, $ret->{dcs}, $ud, $udl );

}

sub parse_status_report($$) {
    my ( $ret, $pdu ) = @_;

    $ret->{reference}  = shift @$pdu;
    $ret->{originator} = parse_address( $pdu, 0 );
    $ret->{sc_stamp}   = parse_timestamp($pdu);
    $ret->{dt_stamp}   = parse_timestamp($pdu);
    $ret->{status_id}  = shift @$pdu;
    $ret->{status}     = $gsm_status{ $ret->{status_id} };
}

sub parse($$) {
    my ( $pdu, $dir ) = @_;
    my $ret = {};

    $ret->{raw} = join( ',', @$pdu );
    $pdu = [@$pdu];

    my $b = shift @{$pdu};

    $ret->{header_len} = $b;

    my $addr = [ splice @{$pdu}, 0, $b ];

    $ret->{smsc} = parse_address( $addr, $b );

    $b = shift @{$pdu};
    $ret->{raw_type_id} = $b & 3;

    unless ( defined $dir ) {
        $dir = SC_TO_ME;
        $dir = SC_TO_ME if $ret->{raw_type_id} == GSM_MTI_DELIVER;
        $dir = ME_TO_SC if $ret->{raw_type_id} == GSM_MTI_SUBMIT;
        $dir = SC_TO_ME if $ret->{raw_type_id} == GSM_MTI_STATUS_REPORT;
    }

    $ret->{direction_id} = $dir;
    $ret->{direction}    = $message_direction[$dir];
    $ret->{type_id}      = $dir + 2 + 2 * $ret->{raw_type_id};
    $ret->{type}         = $message_type[ $ret->{type_id} ];

    $ret->{reject_duplicates}          = ( $b & 0x04 ) ? 1 : 0;
    $ret->{validity_period_type_id}    = ( $b & 0x18 ) >> 3;
    $ret->{status_report_indication}   = ( $b & 0x20 ) ? 1 : 0;
    $ret->{user_data_header_indicator} = ( $b & 0x40 ) ? 1 : 0;
    $ret->{reply_path}                 = ( $b & 0x80 ) ? 1 : 0;

    if ( $ret->{type_id} == MT_DELIVER ) {
        parse_deliver( $ret, $pdu );
    }
    elsif ( $ret->{type_id} == MT_SUBMIT ) {
        parse_submit( $ret, $pdu );
    }
    elsif ( $ret->{type_id} == MT_STATUS_REPORT ) {
        parse_status_report( $ret, $pdu );
    }

    return $ret;
}

sub generate_sms($$$$$$) {
    my ( $dest, $uh, $cc_ref,$part, $total, $text ) = @_;

    my $pdu = [];
    my $udh = [];
    my $ud  = [];

    push @$pdu, 0;    #phone can fill smsc

    my $b = 0x0;

    $b |= GSM_MTI_SUBMIT;
    $b |= 0x10;                   #one byte validity
    $b |= 0x20;                   #always want report
    $b |= 0x40 unless $uh == 0;

    push @$pdu, $b;

#    my $ref=unpack('C',Digest::MD5::md5($text,time()));
    my $ref=1;

    push @$pdu, $ref;              #message ref

    my $a = encode_address($dest);
    push @$pdu, (@$a);

    push @$pdu, 0;                #DCS
    push @$pdu, 0;                #PID

    push @$pdu, 169;              #validity

    my $pad = 0;

    my $udl = length($text);

    if ($uh) {
        my $udhl = 5;

        push @$udh, $udhl;
        push @$udh, GSM_IE_CONCAT_8BIT;
        push @$udh, 3;                    #ie_len
        push @$udh, $cc_ref;
        push @$udh, $total;
        push @$udh, $part;

        my $bits_lost = 8 * ( $udhl + 1 );

        $pad = $bits_lost % 7;
        $pad = 7 - $pad;
        $udl += ( $bits_lost + $pad - 1 ) / 7;
    }

    push @$pdu, $udl;
    push @$pdu, (@$udh) if $uh;

    $ud = pack_7bit_string($text);

    $ud = add_n_bits( $ud, $pad ) if $pad > 0;
    push @$pdu, (@$ud);

    return $pdu;
}

sub generate($$) {
    my ( $dest,  $text ) = @_;
    my $pdus = [];

    $text = encode( "gsm0338", $text );


    if ( length($text) > 160 ) {
        my $t = $text;
        my $total     = 0;
        my $part      = 1;

        my $cc_ref=unpack('C',Digest::MD5::md5($text,time()));

        while ( $t =~ /^(.{154})(.*)/ ) {
            $total++;
            $t = $2;
        }

        $total++ if length($t);

        while ( $text =~ /^(.{154})(.*)/ ) {
            push @$pdus,
              generate_sms( $dest, 1, $cc_ref, $part, $total, $1 );
            $part++;
            $text = $2;
        }
        push @$pdus, generate_sms( $dest, 1,  $cc_ref,$part, $total, $text )
          if length($text);
    }
    else {
        push @$pdus, generate_sms( $dest, 0, 0,0, 0, $text );
    }

    return $pdus;
}

sub array_to_hex($) {
    my $pdu = shift;
    my $ret = "";

    for my $c (@$pdu) {
        $ret .= sprintf "%02X", $c;
    }
    return $ret;
}

sub hex_to_array($) {
    my $pdu = shift;
    my $ret = [];

    while ( $pdu =~ /^(..)(.*)$/ ) {
        push @{$ret}, hex($1);
        $pdu = $2;
    }
    return $ret;
}

1;
