#!/usr/bin/env perl

BEGIN { push @INC, "/etc/xibalba"; }

use strict;
use Data::Dumper;
use Email::Simple;
use MIME::Lite;
use Sys::Syslog qw(:standard :macros);
use GoIP;
use PHX;
use SMS;
use Digest::MD5;
use Asterisk::AMI;

sub send_sms_goip($$$) {
    my ( $ip, $dest, $body ) = @_;

    my $goip = new GoIP( { goip_addr => $ip } );

    syslog( LOG_ERR,
            "attempting via=goip:"
          . $ip
          . " destination="
          . $dest
          . " body="
          . PHX::sanitize($body) );

    my $ok = $goip->sms( $dest, $body );

    if ( ( defined $ok ) and ( $ok == 1 ) ) {
        syslog( LOG_ERR,
                "succeeded  destination="
              . $dest
              . " body="
              . PHX::sanitize($body) );
        return 0;
    }
    else {
        syslog( LOG_ERR,
                "failed     destination="
              . $dest
              . " body="
              . PHX::sanitize($body) );
        return 1;
    }
}

sub send_sms_ami($$$) {
    my ( $device, $pdus, $mid ) = @_;

    my $ami = Asterisk::AMI->new(
        PeerAddr => '127.0.0.1',
        PeerPort => '5038',
        Username => 'ami',
        Secret   => 'ami',
        Events   => 'on'
    );

    my $i = 1;
    my $t = scalar(@$pdus);
    for my $p (@$pdus) {
        my $reference = undef;

	my $pdu=SMS::array_to_hex($p);
        my $resp = $ami->action(
            {
                Action => 'DongleSendPDU',
                Device => $device,
                PDU    => $pdu
            }
        );

        if ( $resp->{Response} ne "Success" ) {
            syslog( LOG_ERR, "-- part $i/$t: DongleSendPDU failed" );
            return 1;
        }

        my $id = $resp->{PARSED}->{ID};

        my $e;
        while ( $e = $ami->get_event() ) {

            $reference = $e->{Reference}
              if ( $e->{Event} eq "DongleSMSReference" )
              and ( $e->{ID} eq $id );

            last
              if ( $e->{Event} eq "DongleSMSStatus" )
              and ( $e->{ID} eq $id );

            # FIXME!! some sort of timeout here
        }

        PHX::store_mid_ref( $mid, $reference ) if defined $reference;
        if ( $t != 1 ) {
            PHX::store_parts_ref( "$i/$t", $reference ) if defined $reference;
        }
        else {
            PHX::store_parts_ref( "", $reference ) if defined $reference;
        }

        syslog( LOG_ERR,
                "-- part $i/$t: DongleSMSStatus "
              . $e->{Status});
	syslog( LOG_ERR,
                "-- part $i/$t: Reference "
              . $reference." PDU ". $pdu );

        return 1 if $e->{Status} ne "Sent";
        $i++;

    }

    return 0;

}

sub get_message() {
    local $/;
    return STDIN->getline;
}

exit 1 unless $ARGV[0] =~ /^([^\.]*)\.(\+*[0-9]*)$/;

my $gate        = $1;
my $destination = $2;

openlog( "xibalba-send.pl " . $gate, "pid", LOG_MAIL );

my $email = Email::Simple->new( get_message() );
my $mid   = $email->header("message-id");
my $body  = $email->body;

$body =~ s/\r\n/\r/sg;
$body =~ s/\n/\r/sg;

$body =~ s/^\r+//s;
$body =~ s/\r+$//s;

exit send_sms_goip( "10.32.87.22", $destination, $body ) if $gate eq "templebar";
exit send_sms_goip( "10.32.87.24", $destination, $body ) if $gate eq "ishtar";

my $device = undef;

$device = "dongle0" if $gate eq "water";
$device = "dongle1" if $gate eq "brandenburg";

exit 1 unless defined $device;

my $ps = SMS::generate( $destination, $body );

syslog( LOG_ERR,
        "attempting via=ami:"
      . $device
      . " destination="
      . $destination
      . " body="
      . PHX::sanitize($body) );

my $ret = send_sms_ami( $device, $ps, $mid );

if ( $ret != 0 ) {
    syslog( LOG_ERR,
            "failed destination="
          . $destination
          . " body="
          . PHX::sanitize($body) );
    exit 1;
}
syslog( LOG_ERR,
    "succeeded destination=" . $destination . " body=" . PHX::sanitize($body) );

exit 0;

